# -*- coding: utf-8 -*-


# import numpy as np
# import warnings


# def interpolate(X: np.array,
#                 Y: np.array,
#                 x: np.array,
#                 flat_extrapolation: bool=True,
#                 log_Y_values: bool=False):
#     """
#     Interpolate the FX forward curve to match given expiry_dates.
#     Please note expiry date is 2 business days prior to the delivery date
#     """
    
#     unique_dates = x.drop_duplicates()
#     combined_index = self.fx_forward_curve.index.union(unique_dates)
#     result = self.fx_forward_curve.reindex(combined_index).copy()
#     start_date, end_date = self.fx_forward_curve.index.min(), self.fx_forward_curve.index.max()
    
#     if flat_extrapolation:
#         try:                    
#             result['fx_forward_rate'] = result['fx_forward_rate'].interpolate(method='time', limit_area='inside').ffill().bfill()
#         except:
#             pass
#         # Find out of range dates and warn
#         out_of_range_dates = unique_dates[(unique_dates < start_date) | (unique_dates > end_date)]
#         for date in out_of_range_dates:
#             warnings.warn(f"Date {date} is outside the range {start_date} - {end_date}, flat extrapolation applied.")
#     else:
#         result['fx_forward_rate'] = result['fx_forward_rate'].interpolate(method='time', limit_area='inside')
    
#     result = result.reindex(x)
    
#     return result['fx_forward_rate']
# Test


# -*- coding: utf-8 -*-
import os
if __name__ == "__main__":
    os.chdir(os.environ.get('PROJECT_DIR_FRM'))

import numpy as np
from numba import jit, prange
import timeit
import time


def simulate_gbm(x0, mu, σ, T, dt, rand_nbs=None, method=0):
    """
    Simulate Geometric Brownian Motion (GBM).

    Parameters:
    x0 (float): Initial value of the process.
    mu (float): Drift.
    σ (float): Volatility.
    T (float): Time end point
    dt (float): Time step size.
    no (np.ndarray, optional): Array of normally distributed pseudorand nbs. Defaults to None.
    method (int, optional): Method for simulation. Defaults to 0.
        0: Direct integration
        1: Euler scheme
        2: Milstein scheme 1st order
        3: Milstein scheme 2nd order

    Returns:
    np.ndarray: Vector containing the simulated trajectory of the GBM.

    References
    [1]: Rouah, F.D. (2013). Euler Scheme for the Black-Scholes Model.
    """

    if rand_nbs is None:
        rand_nbs = np.random.normal(0, 1, int(np.ceil(T / dt)))

    rand_nbs = np.asarray(rand_nbs).flatten()

    if len(rand_nbs) != int(np.ceil(T / dt)):
        raise ValueError("Error: length(no) != n/dt")

    if method == 0:
        x = x0 * np.exp(np.cumsum((mu - 0.5 * σ ** 2) *
                        dt + σ * np.sqrt(dt) * rand_nbs))
    elif method == 1:
        x = x0 * np.cumprod(1 + mu * dt + σ * np.sqrt(dt) * rand_nbs)
    elif method == 2:
        x = x0 * np.cumprod(1 + mu * dt + σ * np.sqrt(dt) * rand_nbs +
                            0.5 * σ ** 2 * dt * (rand_nbs ** 2 - 1))
    elif method == 3:
        x = x0 * np.cumprod(1 + mu * dt + σ * np.sqrt(dt) * rand_nbs +
                            0.5 * σ ** 2 * dt * (rand_nbs ** 2 - 1) +
                            mu * σ * rand_nbs * (dt ** 1.5) + 0.5 * (mu ** 2) * (dt ** 2))

    return np.concatenate(([x0], x))


@jit(nopython=True, parallel=True)
def simulate_gbm_numba_prange(x0, mu, σ, T, dt, rand_nbs=None, method=0):
    steps = int(np.ceil(T / dt))

    if rand_nbs is None:
        rand_nbs = np.random.normal(0, 1, steps)

    rand_nbs = np.asarray(rand_nbs).flatten()

    if len(rand_nbs) != steps:
        raise ValueError("Error: length(no) != n/dt")

    sqrt_dt = np.sqrt(dt)

    x = np.empty(steps)

    if method == 0:
        for i in prange(steps):
            x[i] = np.exp((mu - 0.5 * σ ** 2) * dt + σ * sqrt_dt * rand_nbs[i])
        x = np.cumprod(x)
    elif method == 1:
        for i in prange(steps):
            x[i] = 1 + mu * dt + σ * sqrt_dt * rand_nbs[i]
        x = np.cumprod(x)
    elif method == 2:
        for i in prange(steps):
            x[i] = 1 + mu * dt + σ * sqrt_dt * rand_nbs[i] + \
                0.5 * σ ** 2 * (dt * (rand_nbs[i] ** 2 - 1))
        x = np.cumprod(x)
    elif method == 3:
        for i in prange(steps):
            x[i] = 1 + mu * dt + σ * sqrt_dt * rand_nbs[i] + 0.5 * σ ** 2 * (dt * (rand_nbs[i] ** 2 - 1)) + \
                mu * σ * rand_nbs[i] * (dt ** 1.5) + \
                0.5 * (mu ** 2) * (dt ** 2)
        x = np.cumprod(x)

    result = np.empty(steps + 1)
    result[0] = x0
    for i in prange(steps):
        result[i + 1] = x0 * x[i]

    return result


@jit(nopython=True)
def simulate_gbm_numba(x0, mu, sigma, T, dt, rand_nbs=None, method=0):

    steps = int(np.ceil(T / dt))

    if rand_nbs is None:
        rand_nbs = np.random.normal(0, 1, steps)

    if len(rand_nbs) != steps:
        raise ValueError("Error: length(no) != steps/dt")

    x = np.empty(steps+1)
    x[0] = x0

    if method == 0:
        for i in range(1, steps+1):
            x[i] = x[i-1] * np.exp((mu - 0.5 * sigma ** 2)
                                   * dt + sigma * np.sqrt(dt) * rand_nbs[i-1])
    elif method == 1:  # euler_scheme
        for i in range(1, steps+1):
            x[i] = x[i-1] * (1 + mu * dt + sigma * np.sqrt(dt) * rand_nbs[i-1])
    elif method == 2:  # milstein_scheme_1st_order
        for i in range(1, steps+1):
            x[i] = x[i-1] * (1 + mu * dt + sigma * np.sqrt(dt) * rand_nbs[i-1] +
                             0.5 * sigma ** 2 * dt * (rand_nbs[i-1] ** 2 - 1))
    elif method == 3:  # milstein_scheme_2nd_order
        for i in range(1, steps+1):
            x[i] = x[i-1] * (1 + mu * dt + sigma * np.sqrt(dt) * rand_nbs[i-1] +
                             0.5 * sigma ** 2 * dt * (rand_nbs[i-1] ** 2 - 1) +
                             mu * sigma * rand_nbs[i-1] * (dt ** 1.5) + 0.5 * (mu ** 2) * (dt ** 2))
    return x



if __name__ == "__main__":

    # Numba is much faster after it has been compiled.
    # Key is having a problem where it gets compiled

    # Parameters
    x0 = 100
    mu = 0.05
    σ = 0.2
    n = 10000
    dt = 0.01
    rand_nbs = np.random.normal(0, 1, int(np.ceil(n / dt)))

    t1 = time.time()
    x = simulate_gbm(x0, mu, σ, n, dt, rand_nbs=rand_nbs, method=0)
    t2 = time.time()
    y = simulate_gbm_numba(x0, mu, σ, n, dt, rand_nbs=rand_nbs, method=0)
    t3 = time.time()
    y = simulate_gbm_numba_prange(
        x0, mu, σ, n, dt, rand_nbs=rand_nbs, method=0)
    t4 = time.time()

    print("----------")
    print("GBM:", t2-t1)
    print("GBM numba:", t3-t2)
    print("GBM numba_prange:", t4-t3)

    # Time the original function
    time_original = timeit.timeit(
        'simulate_gbm(x0, mu, σ, n, dt, rand_nbs=rand_nbs, method=0)',
        globals=globals(),
        number=100
    )

    # Time the Numba-optimized function (first call compiles the function)
    simulate_gbm_numba(x0, mu, σ, n, dt, rand_nbs=rand_nbs, method=0)

    time_numba = timeit.timeit(
        'simulate_gbm_numba(x0, mu, σ, n, dt, rand_nbs=rand_nbs, method=0)',
        globals=globals(),
        number=100
    )

    # Time the Numba-optimized function (first call compiles the function)
    simulate_gbm_numba_prange(x0, mu, σ, n, dt, rand_nbs=rand_nbs, method=0)

    time_numba_prange = timeit.timeit(
        'simulate_gbm_numba_prange(x0, mu, σ, n, dt, rand_nbs=rand_nbs, method=0)',
        globals=globals(),
        number=100
    )

    # Display results
    print(f"Time taken by original function: {time_original:.6f} seconds")
    print(f"Time taken by Numba-optimized function: {time_numba:.6f} seconds")
    print(
        f"Time taken by Numba-Prange-optimized function: {time_numba_prange:.6f} seconds")


# -*- coding: utf-8 -*-
import calendar
import datetime as dt
import numpy as np
import pandas as pd
from frm.enums import DayCountBasis


def convert_to_same_shape_datetimeindex(start_date, end_date):
    start_dti = to_datetimeindex(start_date)
    end_dti = to_datetimeindex(end_date)
    
    if len(start_dti) == 1 and len(end_dti) == 1:
        scalar_output = True
    else:
        scalar_output = False
    
    if len(start_dti) == 1 and len(end_dti) > 1:
        start_dti = pd.DatetimeIndex([start_dti.values[0] for _ in range(len(end_dti))])
    elif len(start_dti) > 1 and len(end_dti) == 1:
        end_dti = pd.DatetimeIndex([end_dti.values[0] for _ in range(len(start_dti))])
    
    return start_dti, end_dti, scalar_output


def day_count(start_date,
              end_date, 
              day_count_basis: DayCountBasis,
              is_end_date_on_termination: bool=None)->np.array:
    # If the start_date and end_date are scalars assumption is end_date is the termination date
    # If end_date is a vector, the final value of the vector is assumed to be the termination date
    
    # References
    # [1] The excel file "30-360-2006ISDADefs" sourced from https://www.isda.org/2008/12/22/30-360-day-count-conventions/
    #     Saved to  WayBackMachine on 23 September 2024, https://web.archive.org/web/20240923055727/https://www.isda.org/2008/12/22/30-360-day-count-conventions/
        
    start_dti, end_dti, scalar_output = convert_to_same_shape_datetimeindex(start_date, end_date)
    
    assert (start_dti <= end_dti).all()
    
    if day_count_basis in {DayCountBasis.ACT_360, DayCountBasis.ACT_365, DayCountBasis.ACT_ACT, DayCountBasis.ACT_366}:
        # Act = the actual number of days between the dates
        result = (end_dti - start_dti).days.values
    
    elif day_count_basis == DayCountBasis._30_360:
        # Logic for "30/360" / "360/360" / "Bond Basis" is defined in tab "30-360 Bond Basis" in reference [1]
        
        # If (DAY1=31), Set D1=30, Otherwise set D1=DAY1
        DAY1 = start_dti.day
        d1 = np.where(DAY1 == 31, 30, DAY1) 	
        
        # If (DAY2=31) and (DAY1=30 or 31), Then set D2=30, Otherwise set D2=DAY2	
        DAY2 = end_dti.day
        mask = np.logical_and(d1 == 30, DAY2 == 31)
        d2 = np.where(mask, 30, DAY2)
        
        result = 360*(end_dti.year - start_dti.year) \
               + 30*(end_dti.month - start_dti.month) \
               + d2 - d1
        result = result.values
    
    elif day_count_basis == DayCountBasis._30E_360:
        # Logic for "30/360E" / "Eurobond Basis" is defined in tab "30E-360 Eurobond" in reference [1]
        
        # If (DAY1=31), Set D1=30, Otherwise set D1=DAY1
        DAY1 = start_dti.day
        d1 = np.where(DAY1 == 31, 30, DAY1) 	            

        # If (DAY2=31), Then set D2=30, Otherwise set D2=DAY2	
        DAY2 = end_dti.day
        d2 = np.where(DAY2 == 31, 30, DAY2) 	
        
        result = 360 * (end_dti.year - start_dti.year) \
               + 30 * (end_dti.month - start_dti.month) \
               + d2 - d1
        result = result.values
               
    elif day_count_basis == DayCountBasis._30E_360_ISDA:
        # Logic for 30E/360 (ISDA)" is defined in tab "30E-360 ISDA" in [1]
        
        # Returns TRUE/FALSE if datetime is the last day of February
        def _is_last_day_of_feb(datetime):
            if datetime.month == 2:
                _, last_of_month = calendar.monthrange(datetime.year, datetime.month)
                if datetime.day == last_of_month:
                    return True
            return False

        def _apply_logic(start_date, end_date, is_end_date_on_termination):
            # If (DAY1=31) or (DAY1 is last day of February), Set D1=30, Otherwise set D1=DAY1	
            if start_date.day == 31 or _is_last_day_of_feb(start_date):
                d1 = 30
            else:
                d1 = start_date.day
    
            # If (DAY2=31) or (DAY2 is last day of February but not the Termination Date), Then set D2=30, Otherwise set D2=DAY2
            if end_date.day == 31 or (_is_last_day_of_feb(end_date) and not is_end_date_on_termination):
                d2 = 30
            else:
                d2 = end_date.day
    
            result = 360 * (end_date.year - start_date.year) \
                     + 30 * (end_date.month - start_date.month) \
                     + d2 - d1
            return result
        
        if is_end_date_on_termination is None:
            N = len(end_dti)
            is_end_date_on_termination = [False if i < (N-1)  else True for i in range(N)]
        else:
            is_end_date_on_termination = np.atleast_1d(is_end_date_on_termination)

        result = np.array([_apply_logic(start, end, flag) for start, end, flag in zip(start_dti, end_dti, is_end_date_on_termination)])               
    else:
        raise ValueError
    
    if scalar_output:
        return result.item()
    else:
        return result


def year_frac(start_date,
              end_date,
              day_count_basis: DayCountBasis,
              is_end_date_on_termination: bool=None)->np.array:
    
    # If the start_date and end_date are scalars assumption is end_date is the termination date
    # If end_date is a vector, the final value of the vector is assumed to be the termination date

    if day_count_basis in {DayCountBasis._30_360, DayCountBasis._30E_360, DayCountBasis._30E_360_ISDA, DayCountBasis.ACT_360}:
        return day_count(start_date, end_date, day_count_basis, is_end_date_on_termination) / 360.0
    
    elif day_count_basis == DayCountBasis.ACT_365:
        return day_count(start_date, end_date, day_count_basis, is_end_date_on_termination) / 365.0        

    elif day_count_basis == DayCountBasis.ACT_366:
        return day_count(start_date, end_date, day_count_basis, is_end_date_on_termination) / 366.0 
    
    elif day_count_basis == DayCountBasis.ACT_ACT:
        start_dti, end_dti, scalar_output  = convert_to_same_shape_datetimeindex(start_date, end_date)
        assert (start_dti <= end_dti).all()
        
        start_year = start_dti.year
        end_year = end_dti.year
        year_1_diff = 365 + start_year.map(calendar.isleap)
        year_2_diff = 365 + end_year.map(calendar.isleap) 
        
        total_sum = end_year - start_year - 1
        diff_first = pd.DatetimeIndex([dt.datetime(v + 1, 1, 1) for v in start_year]) - start_dti
        total_sum += diff_first.days / year_1_diff
        diff_second = end_dti - pd.DatetimeIndex([dt.datetime(v, 1, 1) for v in end_year]) 
        total_sum += diff_second.days / year_2_diff

        if scalar_output:
            return total_sum.item()
        else:
            return total_sum
    else:
        raise ValueError


def to_datetimeindex(date_object) -> 'pd.DatetimeIndex':
    """
    Converts a date-like object to a pandas DatetimeIndex.

    Parameters:
    - date_object: Can be of type pd.DatetimeIndex, pd.Timestamp, np.datetime64, dt.date, dt.datetime,
                   pd.Series, or list. Unsupported types will raise a ValueError.

    Returns:
    - pd.DatetimeIndex: The corresponding DatetimeIndex for the provided date-like object.

    Raises:
    - ValueError: If the input type is not supported.

    Supported types:
    - pd.DatetimeIndex: Returned directly.
    - pd.Timestamp: Converted to a DatetimeIndex containing one element.
    - np.datetime64: Converted to a DatetimeIndex containing one element.
    - dt.date / dt.datetime: Converted to a DatetimeIndex containing one element.
    - pd.Series or list: Converted to a DatetimeIndex.
    """        
    if isinstance(date_object, pd.DatetimeIndex):
        return date_object
    if isinstance(date_object, pd.Timestamp):
        return pd.DatetimeIndex([date_object.to_pydatetime()])
    elif isinstance(date_object, np.datetime64):
        return pd.DatetimeIndex([pd.to_datetime(date_object)])
    elif isinstance(date_object, dt.date) or isinstance(date_object, dt.datetime):
        return pd.DatetimeIndex([dt.datetime(date_object.year,date_object.month,date_object.day)])
    elif isinstance(date_object, pd.Series) or isinstance(date_object, list):
        return pd.DatetimeIndex(date_object)
    else:
        raise ValueError("Unsupported type", type(date_object), date_object)
      


if __name__ == "__main__":
    import os
    os.chdir(os.environ.get('PROJECT_DIR_FRM'))
    pass
    # Example usage
    # start_date = pd.date_range(start='2024-01-01', end='2024-12-01', freq='ME')
    # end_date = pd.date_range(start='2024-02-29', end='2024-12-31', freq='ME')
    #
    # day_count_basis = DayCountBasis.from_value('act/365')
    # days = day_count(start_date, end_date, day_count_basis)
    # years = year_frac(start_date, end_date, day_count_basis)
    #
    # for d1, d2, day_count, year_frac in zip(start_date, end_date, days, years):
    #     print(d1.date(), d2.date(), day_count.item(), round(year_frac.item(),6))




# -*- coding: utf-8 -*-
import os
import numpy as np
import pandas as pd
from frm.enums import TermRate, PeriodFreq, DayCountBasis
from frm.utils import tenor_to_date_offset, clean_tenor, get_busdaycal, year_frac, Schedule, workday
from frm.term_structures.zero_curve import ZeroCurve
from frm.pricing_engine.sabr import fit_sabr_params_to_sln_smile
from frm.term_structures.interest_rate_option_helpers import standardise_relative_quote_col_names, standardise_atmf_quote_col_names

if __name__ == "__main__":
    os.chdir(os.environ.get('PROJECT_DIR_FRM'))


fp = 'C:/Users/shasa/Documents/frm_private/tests_private/test_optionlet_support_20240628.xlsm'
#idd_quote_details = pd.read_excel(io=fp, sheet_name='Quotes')
discount_factors_3m = pd.read_excel(io=fp, sheet_name='DF_3M')
discount_factors_6m = pd.read_excel(io=fp, sheet_name='DF_6M')

curve_date = pd.Timestamp('2024-06-28')
cal = get_busdaycal('AUD')
day_count_basis = DayCountBasis.ACT_365

zero_curve_3m = ZeroCurve(curve_date=curve_date,
                          pillar_df=discount_factors_3m,
                          day_count_basis=DayCountBasis.ACT_365,
                          cal=cal,
                          interp_method='linear_on_ln_discount')
zero_curve_6m = ZeroCurve(curve_date=curve_date,
                          pillar_df=discount_factors_6m,
                          day_count_basis=DayCountBasis.ACT_365,
                          cal=cal,
                          interp_method='linear_on_ln_discount')

swaption_quotes = pd.read_excel(io=fp, sheet_name='Swaption1Y')
vol_sln_df = swaption_quotes.loc[swaption_quotes['field']=='lognormal_vol', :].reset_index(drop=True)


fixed_frequency = PeriodFreq.QUARTERLY
settlement_delay = 1
ln_shift = 0.01


# Convert the quote columns to standardised format.
col_name_update_atmf = standardise_atmf_quote_col_names(col_names=list(vol_sln_df.columns))
vol_sln_df.rename(columns=col_name_update_atmf, inplace=True)
col_name_update_relative, col_name_adj_to_forward = standardise_relative_quote_col_names(col_names=list(vol_sln_df.columns))
vol_sln_df.rename(columns=col_name_update_relative, inplace=True)
quote_columns = list(col_name_update_atmf.values()) + list(col_name_update_relative.values())


#%%

df = vol_sln_df[['expiry', 'swap_term', 'frequency']].copy()

# Set the swaption expiry date, and the underlying swap effective date and swap termination date.
df['expiry'] = df['expiry'].apply(clean_tenor)
df['expiry_date'] = workday(curve_date + df['expiry'].apply(tenor_to_date_offset), 0, cal)
df['swap_effective_date'] = workday(df['expiry_date'], settlement_delay, cal)
df['swap_term'] = df['swap_term'].apply(clean_tenor)
df['swap_termination_date'] = workday(df['swap_effective_date'] + df['swap_term'].apply(tenor_to_date_offset), settlement_delay, cal)
df['expiry_years'] = year_frac(curve_date, df['expiry_date'], day_count_basis)

for i, row in df.iterrows():
    df.loc[i,'frequency'] = PeriodFreq.from_value(row['frequency'])


# Functions

# 1. Swaption quote helper - cleans up the swaption quotes dataframe.
# 2. Swaption term structure term rate bootstrapper.
# TODO:
#  Need a 'term rate swap curve'
#  Composed of zero curves and term fixings.
#

# Calculate the par swap rate for each swap term / expiry combo. This is a bit slow (0.2s) for 14 rows.
# Could be optimised by creating the longest schedule over each fixed frequency then indexing.
df['F'] = np.nan
for i, row in df.iterrows():

    freq = row['frequency']

    fixed_schedule = Schedule(start_date=row['swap_effective_date'],
                             end_date=row['swap_termination_date'],
                             freq=freq,
                             cal=cal)
    fixed_schedule.add_period_yearfrac(day_count_basis=day_count_basis)

    if freq == PeriodFreq.QUARTERLY:
        fixed_schedule.df['forward_rate'] = zero_curve_3m.get_forward_rates(period_start=fixed_schedule.df['period_start'],
                                                                            period_end=fixed_schedule.df['period_end'],
                                                                            forward_rate_type=TermRate.SIMPLE)
    elif freq == PeriodFreq.SEMIANNUAL:
        fixed_schedule.df['forward_rate'] = zero_curve_6m.get_forward_rates(period_start=fixed_schedule.df['period_start'],
                                                                            period_end=fixed_schedule.df['period_end'],
                                                                            forward_rate_type=TermRate.SIMPLE)

    fixed_schedule.df['discount_factor'] = zero_curve_3m.get_discount_factors(dates=fixed_schedule.df['payment_date'])
    fixed_schedule.df['annuity_factor'] = fixed_schedule.df['period_yearfrac'] * fixed_schedule.df['discount_factor']
    df.loc[i,'F'] = (fixed_schedule.df['forward_rate'] * fixed_schedule.df['annuity_factor'] \
                     / fixed_schedule.df['annuity_factor'].sum()).sum()

# Setup strikes dataframe
strikes_df = vol_sln_df[quote_columns].copy()
strikes_df[quote_columns] = np.nan
strikes_df.loc[:,'atmf'] = df['F'].values
for col, adj in col_name_adj_to_forward.items():
    strikes_df[col] = strikes_df['atmf'] + adj

#%%

# Each loop is independent, so can be parallelised. 0.27s for 14 rows.
df[['alpha', 'beta', 'rho', 'volvol']] = np.nan
for (i,sabr_row),(j,strikes_row),(k,vol_sln_row) in zip(df.iterrows(),strikes_df.iterrows(),vol_sln_df.iterrows()):
    result = fit_sabr_params_to_sln_smile(tau=sabr_row['expiry_years'],
                                          F=sabr_row['F'],
                                          K=strikes_row[quote_columns],
                                          vols_sln=vol_sln_row[quote_columns],
                                          ln_shift=ln_shift,
                                          beta=vol_sln_row['beta'])

    (alpha, beta, rho, volvol), res = result
    if res.success:
        df.loc[i, ['alpha', 'beta', 'rho', 'volvol']] = alpha, beta, rho, volvol
    else:
        print(res)
        raise ValueError(f"SABR calibration of expiry-swap term {vol_sln_row['expiry']}{vol_sln_row['swap_term']} failed.")

df.to_clipboard()


#%%%



#%%

# Process swaption_quotes into nice format.

# INPUTS
# Dataframe
#  columns names:
#   'expiry': a string of the tenor swaption expiry (e.g. '3M', '1Y', '5Y')
#   'swap_term': a tenor string of the tenor swaption expiry (e.g. '3M', '1Y', '5Y')
#    column names for the strikes of the swaption quotes:
#    (i) atmf and/or
#    (ii) (a) relative quotes to the forward rate (e.g. '-100bps', '-50bps', '0bps', '50bps', '100bps') or
#         (b) absolute quotes (e.g. '2.5%', '3.0%', '3.5%')
#         Note only (a) or (b) can be specified, not (a) and (b) together.
#   fixed leg frequency, day count basis
#   floating leg frequency, day count basis
# curve_date
# busdaycal
# roll_convention
# zero_curve
# swap_settle_delay_after_expiry
# settlement_date

# To do
# Format the swaption quotes into a nice format
# validate, settlement date, settlement delay
# calculate the forward rate for each swap term / expiry combo (adjust for the fixed/float rate day count & freq)
# calculate the expiry, effective date, termination date for swap term / expiry combo
# calculate the annuity factor for each swap term / expiry combo (if needed).

# SABR fit function.
# Fit sabr params to the swaption quotes.

# Interp function

# Interp term structure for any term and expiry. Linear interp on beta, rho, volvol and vol_n_atmf.
# alpha and vol_sln_atmf are analytically calculated.

# Swaption pricer for any swaption.
# Given annuity factor is the dimension of variation of schedule, requiring a fixed/float schedule object/df would be easy.
# THis is a later priority, once we map how to define swap legs and a trade/swap object (Pay/Rcv leg is where I am thinking currently).



# -*- coding: utf-8 -*-
import os
if __name__ == "__main__":
    os.chdir(os.environ.get('PROJECT_DIR_FRM'))
import numpy as np
import pandas as pd
import re

from frm.utils.daycount import year_frac
from frm.enums.utils import DayCountBasis, PeriodFrequency
from frm.enums.term_structures import TermRate
from frm.utils.tenor import clean_tenor, tenor_to_date_offset
from frm.utils.utilities import convert_column_to_consistent_data_type
from frm.utils.schedule import get_schedule
from frm.term_structures.zero_curve import ZeroCurve

from typing import Optional






def process_capfloor_quotes(
        curve_date: pd.Timestamp,
        vol_ln_df: pd.DataFrame,
        ln_shift: float,
        day_count_basis: DayCountBasis,
        zero_curve: ZeroCurve,
        optionlet_date_offset: pd.DateOffset,
        settlement_date: Optional[pd.Timestamp]=None,
        settlement_delay: Optional[int]=1,
        busdaycal: Optional[np.busdaycalendar]=np.busdaycalendar()
):

    if settlement_date is None:
        settlement_date = np.busday_offset(curve_date.to_numpy().astype('datetime64[D]'), offsets=settlement_delay, roll='following', busdaycal=busdaycal)

    effective_date_np = (settlement_date + optionlet_date_offset).to_numpy().astype('datetime64[D]')
    effective_date = np.busday_offset(effective_date_np, offsets=0, roll='following', busdaycal=busdaycal)

    vol_ln_df['tenor'] = vol_ln_df['tenor'].apply(clean_tenor)
    vol_ln_df['settlement_date'] = settlement_date
    vol_ln_df['effective_date'] = effective_date

    for i, row in vol_ln_df.iterrows():
        date_offset = tenor_to_date_offset(row['tenor'])
        termination_date = settlement_date + date_offset
        last_optionlet_expiry_date = termination_date - optionlet_date_offset
        last_optionlet_expiry_date_np = last_optionlet_expiry_date.to_numpy().astype('datetime64[D]')
        termination_date_date_np = termination_date.to_numpy().astype('datetime64[D]')
        vol_ln_df.at[i, 'last_optionlet_expiry_date'] = np.busday_offset(last_optionlet_expiry_date_np, offsets=0, roll='following',busdaycal=busdaycal)
        vol_ln_df.at[i, 'termination_date'] = np.busday_offset(termination_date_date_np, offsets=0, roll='following', busdaycal=busdaycal)

    vol_ln_df['term_years'] = year_frac(vol_ln_df['effective_date'], vol_ln_df['termination_date'], day_count_basis)
    vol_ln_df['last_optionlet_expiry_years'] = year_frac(curve_date, vol_ln_df['last_optionlet_expiry_date'], day_count_basis)
    vol_ln_df['F'] = np.nan
    vol_ln_df['ln_shift'] = ln_shift

    vol_ln_df = convert_column_to_consistent_data_type(vol_ln_df)

    first_columns = ['tenor', 'settlement_date', 'effective_date', 'last_optionlet_expiry_date','termination_date', 'term_years','last_optionlet_expiry_years','F','ln_shift']
    column_order = first_columns + [col for col in vol_ln_df.columns if col not in first_columns]
    vol_ln_df = vol_ln_df[column_order]

    # Two methods of specifying quotes, same functions can be used for Caps/Floors and Swaptions
    # 1. Quotes relative to the atm forward rate (e.g. ATM, ATM+/-50bps, ATM+/-100bps...)
    # 2. Absolute quotes (e.g. 2.5%, 3.0%, 3.5%...)

    col_name_update_dict, quote_str_map = standardise_relative_quote_col_names(col_names=list(vol_ln_df.columns))
    vol_ln_df.rename(columns=col_name_update_dict, inplace=True)
    vol_ln_df.reset_index(drop=True, inplace=True)

    N = len(vol_ln_df) - 1
    effective_date = vol_ln_df.loc[N,'effective_date']
    termination_date = vol_ln_df.loc[N,'termination_date']
    optionlet_df = get_schedule(start_date=effective_date,
                                end_date=termination_date,
                                frequency=PeriodFrequency.QUARTERLY,
                                busdaycal=busdaycal)
    optionlet_df['coupon_term'] = year_frac(optionlet_df['period_start'], optionlet_df['period_end'], day_count_basis)
    optionlet_df['discount_factors'] = zero_curve.get_discount_factors(dates=optionlet_df['payment_dates'])
    optionlet_df['annuity_factor'] = optionlet_df['coupon_term'] * optionlet_df['discount_factors']
    optionlet_df['expiry_years'] = year_frac(curve_date, optionlet_df['period_start'], day_count_basis)
    optionlet_df['F'] = zero_curve.get_forward_rates(period_start=optionlet_df['period_start'],
                                                     period_end=optionlet_df['period_end'],
                                                     forward_rate_type=TermRate.SIMPLE)
    optionlet_df['vol_n_atm'] = np.nan
    optionlet_df['vol_sln_atm'] = np.nan
    optionlet_df['ln_shift'] = ln_shift
    optionlet_df['alpha'] = np.nan
    optionlet_df['beta'] = np.nan
    optionlet_df['rho'] = np.nan
    optionlet_df['volvol'] = np.nan

    # Calculate the forward rate (pre lognormal shift) for the cap/floor quotes
    for i,row in vol_ln_df.iterrows():
        mask = (optionlet_df['period_end'] <= vol_ln_df.loc[i,'termination_date'])
        vol_ln_df.loc[i,'F'] = (optionlet_df.loc[mask, 'F'] * optionlet_df.loc[mask, 'annuity_factor']).sum() / optionlet_df.loc[mask, 'annuity_factor'].sum()

    # Setup a strike dataframe
    strikes_df = vol_ln_df.copy()
    strikes_df[list(quote_str_map.keys())] = np.nan
    for column in strikes_df[quote_str_map.keys()].columns:
        strikes_df[column] = strikes_df['F'] + quote_str_map[column]

    # Setup a call/put flag dataframe
    call_put_df = vol_ln_df.copy()
    call_put_df[list(quote_str_map.keys())] = np.nan
    call_put_df[list(quote_str_map.keys())] = np.where(strikes_df[list(quote_str_map.keys())].values > call_put_df['F'].values[:, None], 1, -1)


    capfloor_surf = dict()
    capfloor_surf['vol_ln_df'] = vol_ln_df
    capfloor_surf['strikes'] = strikes_df
    capfloor_surf['call_put_df'] = call_put_df
    capfloor_surf['optionlet_df'] = optionlet_df
    capfloor_surf['zero_curve'] = zero_curve
    capfloor_surf['quote_columns'] = list(quote_str_map.keys())

    return capfloor_surf



def settlement_date_helper(curve_date: pd.Timestamp,
                           settlement_date: Optional[pd.Timestamp],
                           settlement_delay: Optional[int],
                           busdaycal: Optional[np.busdaycalendar]=np.busdaycalendar()):

        if settlement_date is None and settlement_delay is None:
            raise ValueError('Either settlement_date or settlement_delay must be provided.')

        elif settlement_date is None and settlement_delay is not None:
            settlement_date = np.busday_offset(curve_date.to_numpy().astype('datetime64[D]'),
                                                    offsets=settlement_delay,
                                                    roll='following',
                                                    busdaycal=busdaycal)
        elif settlement_date is not None and settlement_delay is None:
            assert settlement_date >= curve_date, f"settlement_date {settlement_date} must be on or after curve_date {curve_date}."

        return settlement_date


def standardise_atmf_quote_col_names(col_names: list[str]):
    col_name_update = dict()
    atm_quote = '(a|at)[ -]?(t|the)[ -]?(m|money)[ -]?(f|forward)?'

    for col_name in col_names:
        if re.search(atm_quote, col_name, re.IGNORECASE):
            new_col_name = 'atmf'
            col_name_update[col_name] = new_col_name

    assert len(col_name_update) == 1, f"Expected 1 ATMF quote, found {len(col_name_update)}."

    return col_name_update


def standardise_relative_quote_col_names(col_names: list[str]):
    col_name_update = dict()
    col_name_adj_to_forward = dict()
    bps_quote = r'[+-]?\s?\d+\s?(bps|bp)'
    percentage_quote = r'[+-]?\s?\d+(\.\d+)?\s?%'

    # ATMF quote
    col_name_update_atmf = standardise_atmf_quote_col_names(col_names=col_names)
    col_name_update.update(col_name_update_atmf)
    col_name_adj_to_forward.update({'atmf': 0})

    # Relative Quotes
    for col_name in col_names:
        if re.search(bps_quote, col_name, re.IGNORECASE):
            v = round(float(col_name.replace('bps', '').replace('bp', '').replace(' ', '')) / 10000,8)
            new_col_name = (str(int(v * 10000)) if (round(v * 10000,8)).is_integer() else str(round(v * 10000, 8))) + 'bps'
            col_name_update[col_name] = new_col_name
            col_name_adj_to_forward[new_col_name] = v

        elif re.search(percentage_quote, col_name, re.IGNORECASE):
            v = round(float(col_name.replace('%', '').replace(' ', '')) / 100,8)
            new_col_name = (str(int(v * 100)) if (round(v * 100,8)).is_integer() else str(round(v * 100, 8))) + 'bps'
            col_name_update[col_name] = new_col_name
            col_name_adj_to_forward[new_col_name] = v

    return col_name_update, col_name_adj_to_forward


def standardise_absolute_quote_col_names(col_names: list[str]):
    col_name_update = dict()
    col_name_strike = dict()
    bps_quote = r'[+-]?\s?\d+\s?(bps|bp)'
    percentage_quote = r'[+-]?\s?\d+(\.\d+)?\s?%'

    for col_name in col_names:
        if re.search(bps_quote, col_name, re.IGNORECASE):
            v = round(float(col_name.replace('bps', '').replace('bp', '').replace(' ', '')) / 10000,8)
            new_col_name = (str(int(v * 10000)) if (round(v * 10000,8)).is_integer() else str(round(v * 10000, 8))) + 'bps'
            col_name_update[col_name] = new_col_name
            col_name_strike[new_col_name] = v

        elif re.search(percentage_quote, col_name, re.IGNORECASE):
            v = round(float(col_name.replace('%', '').replace(' ', '')) / 100,8)
            new_col_name = (str(int(v * 100)) if (round(v * 100,8)).is_integer() else str(round(v * 100, 8))) + 'bps'
            col_name_update[col_name] = new_col_name
            col_name_strike[new_col_name] = v

    return col_name_update, col_name_strike


# -*- coding: utf-8 -*-
import os

from frm.utils import workday

if __name__ == "__main__":
    os.chdir(os.environ.get('PROJECT_DIR_FRM')) 

from frm.enums.term_structures import FXSmileInterpolationMethod
from frm.term_structures.zero_curve import ZeroCurve
from frm.pricing_engine.monte_carlo_generic import generate_rand_nbs
from frm.pricing_engine.heston import heston_calibrate_vanilla_smile, heston_price_vanilla_european, heston_simulate
from frm.pricing_engine.garman_kohlhagen import garman_kohlhagen_price, garman_kohlhagen_solve_strike_from_delta, garman_kohlhagen_solve_implied_vol
from frm.pricing_engine.geometric_brownian_motion import simulate_gbm_path
from frm.pricing_engine.volatility_generic import forward_volatility, flat_forward_interp
from frm.term_structures.fx_helpers import (clean_vol_quotes_column_names,
                                            get_delta_smile_quote_details,
                                            check_delta_convention,
                                            fx_term_structure_helper,
                                            fx_forward_curve_helper,
                                            interp_fx_forward_curve_df,
                                            validate_ccy_pair,
                                            solve_call_put_quotes_from_strategy_quotes)

from scipy.interpolate import CubicSpline, InterpolatedUnivariateSpline
from frm.utils import year_frac, get_busdaycal, resolve_fx_curve_dates
from frm.enums import DayCountBasis, CompoundingFreq

import numpy as np
import pandas as pd
from dataclasses import dataclass, field, InitVar
from typing import Optional
import matplotlib.pyplot as plt
import matplotlib.ticker as mtick


@dataclass
class FXVolatilitySurface:
    # Mandatory initialisation attributes
    domestic_ccy: str
    foreign_ccy: str
    fx_forward_curve_df: pd.DataFrame # Quotes in perspective of # of domestic currency units per foreign currency unit
    domestic_zero_curve: ZeroCurve
    foreign_zero_curve: ZeroCurve
    vol_quotes: InitVar[pd.DataFrame]

    #### Optional initialisation attributes

    # One of curve_date or spot_date must be provided. If spot_date is not provided, it will be calculated from curve_date and spot_offset or set per market convention
    curve_date: Optional[pd.Timestamp] = None
    spot_offset: Optional[int] = None
    spot_date: Optional[pd.Timestamp] = None

    cal: np.busdaycalendar = None
    day_count_basis: DayCountBasis = DayCountBasis.ACT_ACT
    smile_interpolation_method: FXSmileInterpolationMethod = FXSmileInterpolationMethod.UNIVARIATE_SPLINE

    # Non initialisation attributes set in __post_init__
    fx_spot_rate: float = field(init=False)
    vol_smile_pillar_df: pd.DataFrame = field(init=False)
    vol_smile_daily_df: pd.DataFrame = field(init=False)
    vol_smile_daily_func: dict = field(init=False)

    quotes_column_names: np.ndarray = field(init=False)
    quotes_signed_delta: np.ndarray = field(init=False)
    quotes_call_put_flag: np.ndarray = field(init=False)


    def __post_init__(self, vol_quotes):

        self.vol_smile_daily_func = {}

        self.ccy_pair, self.domestic_ccy, self.foreign_ccy =  validate_ccy_pair(self.foreign_ccy + self.domestic_ccy)

        if self.cal is None:
            self.cal = get_busdaycal([self.domestic_ccy, self.foreign_ccy])

        self.curve_date, self.spot_offset, self.spot_date = resolve_fx_curve_dates(self.ccy_pair,
                                                                                   self.cal,
                                                                                   self.curve_date,
                                                                                   self.spot_offset,
                                                                                   self.spot_date)
        assert self.curve_date == self.domestic_zero_curve.curve_date
        assert self.curve_date == self.foreign_zero_curve.curve_date

        self.fx_forward_curve_df = fx_forward_curve_helper(fx_forward_curve_df=self.fx_forward_curve_df,
                                                           curve_date=self.curve_date,
                                                           spot_offset=self.spot_offset,
                                                           cal=self.cal)
        mask = self.fx_forward_curve_df['tenor'] == 'sp'
        self.fx_spot_rate = self.fx_forward_curve_df.loc[mask, 'fx_forward_rate'].values[0]

        # Process volatility input and setup pillar dataframe and daily helper dataframe
        vol_quotes = clean_vol_quotes_column_names(vol_quotes)
        vol_quotes_call_put = solve_call_put_quotes_from_strategy_quotes(vol_quotes)
        delta_smile_quote_details = get_delta_smile_quote_details(vol_quotes_call_put)
        self.quotes_column_names = np.array(delta_smile_quote_details['quotes_column_names'].to_list())
        self.quotes_signed_delta = np.array(delta_smile_quote_details['quotes_signed_delta'].to_list())
        self.quotes_call_put_flag = np.array(delta_smile_quote_details['quotes_call_put_flag'].to_list())
        self.vol_smile_pillar_df = self._setup_vol_smile_pillar_df(vol_quotes_call_put)
        self.vol_smile_daily_df = self._setup_vol_smile_daily_df()
        self.strike_pillar_df = self._setup_strike_pillar()


    def _add_fx_ir_to_df(self, df: pd.DataFrame) -> pd.DataFrame:
        df['fx_forward_rate'] = interp_fx_forward_curve_df(fx_forward_curve_df=self.fx_forward_curve_df,
                                                           dates=df['expiry_date'],
                                                           date_type='fixing_date')
        df['domestic_zero_rate'] = self.domestic_zero_curve.get_zero_rates(dates=df['expiry_date'],
                                                                          compounding_freq=CompoundingFreq.CONTINUOUS)
        df['foreign_zero_rate'] = self.foreign_zero_curve.get_zero_rates(dates=df['expiry_date'],
                                                                        compounding_freq=CompoundingFreq.CONTINUOUS)
        return df


    def _setup_vol_smile_pillar_df(self, vol_smile_pillar_df):
            vol_smile_pillar_df['warnings'] = ''

            vol_smile_pillar_df = check_delta_convention(vol_smile_pillar_df, self.ccy_pair)
            vol_smile_pillar_df = fx_term_structure_helper(df=vol_smile_pillar_df,
                                                           curve_date=self.curve_date,
                                                           spot_offset=self.spot_offset,
                                                           cal=self.cal,
                                                           rate_set_date_str='expiry_date')
            vol_smile_pillar_df['expiry_years'] = year_frac(self.curve_date, vol_smile_pillar_df['expiry_date'], self.day_count_basis)
            vol_smile_pillar_df = self._add_fx_ir_to_df(vol_smile_pillar_df)
            column_order = ['warnings'] \
                           + (['tenor'] if 'tenor' in vol_smile_pillar_df.columns else []) \
                           + ['expiry_date', 'delivery_date', 'expiry_years', 'fx_forward_rate', 'domestic_zero_rate',
                              'foreign_zero_rate', 'delta_convention'] \
                           + list(self.quotes_column_names)
            return vol_smile_pillar_df[column_order]


    def _setup_vol_smile_daily_df(self):
        min_expiry = self.vol_smile_pillar_df['expiry_date'].min()
        max_expiry = self.vol_smile_pillar_df['expiry_date'].max()
        expiry_dates = pd.date_range(min_expiry, max_expiry, freq='d')
        delivery_dates = workday(expiry_dates, offset=self.spot_offset, cal=self.cal)

        expiry_years = year_frac(self.curve_date, expiry_dates, self.day_count_basis)

        vol_smile_daily_df = pd.DataFrame({'expiry_date_daily': expiry_dates,
                                           'delivery_date_daily': delivery_dates,
                                           'expiry_years_daily': expiry_years})

        vol_smile_pillar_df = self.vol_smile_pillar_df.copy()
        vol_smile_pillar_df['expiry_years'] = year_frac(self.curve_date, vol_smile_pillar_df['expiry_date'],self.day_count_basis)

        # Merge to find closest smaller and larger tenors for each target tenor
        lower_pillar = pd.merge_asof(vol_smile_daily_df, vol_smile_pillar_df, left_on='expiry_date_daily',
                                     right_on='expiry_date', direction='backward')
        upper_pillar = pd.merge_asof(vol_smile_daily_df, vol_smile_pillar_df, left_on='expiry_date_daily',
                                     right_on='expiry_date', direction='forward')

        # TODO: Adjust the interpolation to consider the delta convention where pillar points have different delta conventions
        vol_smile_daily_df['delta_convention'] = upper_pillar['delta_convention']

        # Convert to numpy for efficient calculations
        t1 = lower_pillar['expiry_years'].to_numpy()
        t2 = upper_pillar['expiry_years'].to_numpy()
        t = vol_smile_daily_df['expiry_years_daily'].to_numpy()
        t1 = t1[:, np.newaxis]
        t2 = t2[:, np.newaxis]
        t = t[:, np.newaxis]
        vol_t1 = lower_pillar[self.quotes_column_names].to_numpy()
        vol_t2 = upper_pillar[self.quotes_column_names].to_numpy()

        vol_smile_daily_df.rename(columns={'expiry_date_daily': 'expiry_date',
                                           'expiry_years_daily': 'expiry_years'}, inplace=True)

        # Interpolate the volatility smile
        vol_smile_daily_df[self.quotes_column_names] = flat_forward_interp(t1, vol_t1, t2, vol_t2, t)

        vol_smile_daily_df = self._add_fx_ir_to_df(vol_smile_daily_df)

        column_order = ['expiry_date', 'expiry_years', 'fx_forward_rate', 'domestic_zero_rate', 'foreign_zero_rate', 'delta_convention'] + list(self.quotes_column_names)

        return vol_smile_daily_df[column_order]


    def _setup_strike_pillar(self):
        strike_pillar_df = self.vol_smile_pillar_df.copy()
        strike_pillar_df.loc[:, self.quotes_column_names] = np.nan
        for i, row in strike_pillar_df.iterrows():
            strikes = garman_kohlhagen_solve_strike_from_delta(
                S0=self.fx_spot_rate,
                tau=row['expiry_years'],
                r_d=row['domestic_zero_rate'],
                r_f=row['foreign_zero_rate'],
                vol=self.vol_smile_pillar_df.loc[i, self.quotes_column_names].values,
                signed_delta=self.quotes_signed_delta,
                delta_convention=row['delta_convention'],
                F=row['fx_forward_rate'])
            strike_pillar_df.loc[i, self.quotes_column_names] = strikes
        return strike_pillar_df


    def _solve_vol_daily_smile_func(self, expiry_dates: pd.DatetimeIndex):
        """
        Internal method for solving the function definition for the volatility smile.
        The 'function' is a spline or the Heston parameters that allow for the interpolation of the volatility smile for any strike/delta.
        The function is solved for each date in `expiry_dates` and stored in the `vol_smile_daily_func` attribute.

        Parameters:
        dates (pd.DatetimeIndex): Array of dates to perform the σ surface interpolation for.

        Raises:
        ValueError: If the sum of squared errors (SSE) from the Heston fit exceeds a threshold, indicating a poor fit.
        """

        for expiry_date in expiry_dates.unique():
            if expiry_date not in self.vol_smile_daily_func.keys():

                mask = self.vol_smile_daily_df['expiry_date'] == expiry_date
                row = self.vol_smile_daily_df.loc[mask].copy()

                # Scalars
                S0 = self.fx_spot_rate
                r_f=row['foreign_zero_rate'].iloc[0]
                r_d=row['domestic_zero_rate'].iloc[0]
                tau=row['expiry_years'].iloc[0]
                delta_convention=row['delta_convention'].iloc[0]
                F=row['fx_forward_rate'].iloc[0]

                if F is not None:
                    # Use market forward rate and imply the curry basis-adjusted domestic interest rate
                    F = np.atleast_1d(F).astype(float)
                    r_d_basis_adj = np.log(F / S0) / tau + r_f  # from F = S0 * exp((r_d - r_f) * tau)
                    r = r_d_basis_adj
                    q = r_f
                else:
                    r = r_d
                    q = r_f

                # Arrays
                vol = row[self.quotes_column_names].iloc[0].values
                signed_delta = self.quotes_signed_delta
                cp = self.quotes_call_put_flag

                if self.smile_interpolation_method in [FXSmileInterpolationMethod.UNIVARIATE_SPLINE,
                                                       FXSmileInterpolationMethod.CUBIC_SPLINE]:
                    K = garman_kohlhagen_solve_strike_from_delta(
                        S0=S0, tau=tau, r_d=r_d, r_f=r_f, vol=vol, signed_delta=signed_delta, delta_convention=delta_convention, F=F)
                    if self.smile_interpolation_method == FXSmileInterpolationMethod.UNIVARIATE_SPLINE:
                        if len(self.quotes_column_names) < 3:
                            raise ValueError('Cannot fit InterpolatedUnivariateSpline with less than 3 points, please provide more points.')
                        degree = min(3, max(2, len(self.quotes_column_names) - 1))
                        self.vol_smile_daily_func[expiry_date] = InterpolatedUnivariateSpline(x=K, y=vol, k=degree)
                    elif self.smile_interpolation_method == FXSmileInterpolationMethod.CUBIC_SPLINE:
                        if len(self.quotes_column_names) < 4:
                            raise ValueError('Cannot fit CubicSpline with less than 4 points, please provide more points or use a different interpolation method')
                        self.vol_smile_daily_func[expiry_date] = CubicSpline(x=K, y=vol)
                elif self.smile_interpolation_method in [FXSmileInterpolationMethod.HESTON_1993,
                                                         FXSmileInterpolationMethod.HESTON_CARR_MADAN_GAUSS_KRONROD_QUADRATURE,
                                                         FXSmileInterpolationMethod.HESTON_CARR_MADAN_FFT_W_SIMPSONS,
                                                         FXSmileInterpolationMethod.HESTON_COSINE,
                                                         FXSmileInterpolationMethod.HESTON_LIPTON]:
                    var0, vv, kappa, theta, rho, lambda_, IV, SSE = \
                        heston_calibrate_vanilla_smile(volatility_quotes=vol,
                                                       delta_of_quotes=signed_delta,
                                                       S0=S0,
                                                       r=r,
                                                       q=q,
                                                       tau=tau,
                                                       cp=cp,
                                                       delta_convention=delta_convention,
                                                       pricing_method=self.smile_interpolation_method.value)

                    if SSE < 0.001:
                        result = {
                            'var0': var0,
                            'vv': vv,
                            'kappa': kappa,
                            'theta': theta,
                            'rho': rho,
                            'lambda_': lambda_,
                            'IV': IV,
                            'SSE': SSE
                        }
                        self.vol_smile_daily_func[expiry_date] = result
                    else:
                        raise ValueError('SSE is a large value, ', round(SSE, 4), ' heston fit at ', expiry_date,
                                         ' is likely poor')


    def interp_vol_surface(self,
                         expiry_dates: pd.DatetimeIndex,
                         K: np.ndarray,
                         cp: np.ndarray) -> pd.DataFrame:

        K = np.atleast_1d(K).astype(float)
        cp = np.atleast_1d(cp).astype(float)
        assert expiry_dates.shape == K.shape
        assert expiry_dates.shape == cp.shape

        self._solve_vol_daily_smile_func(expiry_dates)

        mask = self.vol_smile_daily_df['expiry_date'].isin(expiry_dates)
        tmp_df = self.vol_smile_daily_df.loc[mask].copy()
        tmp_df.set_index('expiry_date', inplace=True)
        interp_df = tmp_df.loc[expiry_dates].copy()
        interp_df.index.name = 'expiry_date'
        interp_df.reset_index(inplace=True, drop=False)
        interp_df['K'] = K
        interp_df['cp'] = cp
        interp_df['vol'] = np.nan

        for i,row in interp_df.iterrows():
            vol_smile_func = self.vol_smile_daily_func[row['expiry_date']]

            if self.smile_interpolation_method in [FXSmileInterpolationMethod.UNIVARIATE_SPLINE,
                                                    FXSmileInterpolationMethod.CUBIC_SPLINE]:
                interp_df.loc[i,'vol'] = vol_smile_func(K[i])
            elif self.smile_interpolation_method in [FXSmileInterpolationMethod.HESTON_1993,
                                                    FXSmileInterpolationMethod.HESTON_CARR_MADAN_GAUSS_KRONROD_QUADRATURE,
                                                    FXSmileInterpolationMethod.HESTON_CARR_MADAN_FFT_W_SIMPSONS,
                                                    FXSmileInterpolationMethod.HESTON_COSINE,
                                                    FXSmileInterpolationMethod.HESTON_LIPTON]:

                S0 = self.fx_spot_rate
                r_f = row['foreign_zero_rate']
                r_d = row['domestic_zero_rate']
                tau = row['expiry_years']
                F = row['fx_forward_rate']

                if F is not None:
                    # Use market forward rate and imply the curry basis-adjusted domestic interest rate
                    F = np.atleast_1d(F).astype(float)
                    r_d_basis_adj = np.log(F / S0) / tau + r_f  # from F = S0 * exp((r_d - r_f) * tau)
                    r = r_d_basis_adj
                    q = r_f
                else:
                    r = r_d
                    q = r_f

                X = heston_price_vanilla_european(
                    S0=S0, tau=tau, r=r, q=q, cp=cp[i], K=K[i], var0=vol_smile_func['var0'],
                    vv=vol_smile_func['vv'], kappa=vol_smile_func['kappa'], theta=vol_smile_func['theta'],
                    rho=vol_smile_func['rho'], lambda_=vol_smile_func['lambda_'],
                    pricing_method=self.smile_interpolation_method.value)

                interp_df.loc[i,'vol'] = garman_kohlhagen_solve_implied_vol(
                    S0=S0, tau=tau, r_d=r, r_f=q, cp=cp[i], K=K[i], X=X,
                    vol_guess=np.sqrt(vol_smile_func['var0']))

        return interp_df


    def price_vanilla_european(self,
                               expiry_dates: pd.DatetimeIndex,
                               K: [float, np.ndarray],
                               cp: [float, np.ndarray],
                               analytical_greeks_flag: bool=False,
                               numerical_greeks_flag: bool=False,
                               intrinsic_time_split_flag: bool=False) -> dict:

        K, cp = map(np.atleast_1d, (K, cp))
        assert expiry_dates.shape == K.shape
        assert expiry_dates.shape == cp.shape

        interp_df = self.interp_vol_surface(expiry_dates=expiry_dates, K=K, cp=cp)

        results = garman_kohlhagen_price(
            S0=self.fx_spot_rate,
            tau=interp_df['expiry_years'].values,
            r_d=interp_df['domestic_zero_rate'].values,
            r_f=interp_df['foreign_zero_rate'].values,
            cp=cp,
            K=K,
            vol=interp_df['vol'].values,
            F=interp_df['fx_forward_rate'].values,
            analytical_greeks=analytical_greeks_flag,
            numerical_greeks=numerical_greeks_flag,
            intrinsic_time_split=intrinsic_time_split_flag
        )

        results['market_data_inputs'] = pd.DataFrame({'vol': interp_df['vol'].values,
                                                      'r_d': interp_df['domestic_zero_rate'].values,
                                                      'r_f': interp_df['foreign_zero_rate'].values,
                                                      'F': interp_df['fx_forward_rate'].values})

        return results


    def simulate_gbm_fx_rate_path(self,
                                  delivery_date_grid: pd.DatetimeIndex,
                                  nb_simulations: Optional[int]=None,
                                  apply_antithetic_variates: bool=True):

        delivery_date_grid = delivery_date_grid.unique().sort_values(ascending=True)
        delivery_date_grid = delivery_date_grid.union(self.spot_date)
        fixing_date_grid = np.busday_offset(delivery_date_grid, offsets=-1*self.spot_offset, roll='preceding', busdaycal=self.cal)
        mask = self.vol_smile_daily_df['expiry_date'].isin(fixing_date_grid)
        vol_smile_daily_df = self.vol_smile_daily_df.loc[mask].copy()

        schedule = pd.DataFrame({
            'fixing_date': fixing_date_grid,
            'delivery_date': delivery_date_grid,
            'fixing_years': year_frac(self.curve_date, delivery_date_grid, self.day_count_basis).values,
            'domestic_zero_rates': vol_smile_daily_df['domestic_zero_rate'].values,
            'foreign_zero_rates': vol_smile_daily_df['foreign_zero_rate'].values,
            'fx_forward_rates:': vol_smile_daily_df['fx_forward_rate'].values,
        })

        if 'atm_forward' in self.vol_smile_daily_df.columns:
            schedule['atm_forward'] = vol_smile_daily_df['atm_forward'].values
        elif 'atm_delta_neutral' in self.vol_smile_daily_df.columns:
            schedule['volatility'] = vol_smile_daily_df['atm_delta_neutral'].values
        else:
            raise ValueError

        schedule['dt'] = schedule['fixing_years'].diff()
        schedule['dt'].iloc[0] = 0.0

        schedule['drift'] = np.log(schedule['fx_forward_rates:'].shift(-1) / schedule['fx_forward_rates:']) / schedule['dt']
        schedule['drift'].iloc[0] = np.nan

        schedule['forward_volatility'] = forward_volatility(
            t1=schedule['fixing_years'].values[:-1],
            vol_t1=schedule['volatility'].values[:-1],
            t2=schedule['fixing_years'].values[1:],
            vol_t2=schedule['volatility'].values[1:]
        )

        results = {'gbm_monte_carlo_market_data_inputs': schedule}

        rand_nbs = generate_rand_nbs(nb_steps=len(delivery_date_grid) - 1,
                                     nb_rand_vars=1,
                                     nb_simulations=nb_simulations,
                                     apply_antithetic_variates=apply_antithetic_variates)

        fx_rate_simulation_paths = simulate_gbm_path(initial_px=self.fx_spot_rate,
                                                     drift=schedule['drift'].values[1:],
                                                     forward_volatility=schedule['forward_volatility'].values[1:],
                                                     timestep_length=schedule['dt'].values[1:],
                                                     rand_nbs=rand_nbs)

        results['fx_rate_simulation_paths'] = fx_rate_simulation_paths

        return results


    def simulate_heston_fx_rate_path(self,
                                     delivery_date_grid,
                                     nb_simulations=None,
                                     flag_apply_antithetic_variates=True):


        delivery_date_grid = delivery_date_grid.unique().sort_values(ascending=True)
        delivery_date_grid = delivery_date_grid.union(self.spot_date)
        fixing_date_grid = np.busday_offset(delivery_date_grid, offsets=-1*self.spot_offset, roll='preceding', busdaycal=self.cal)
        #mask = self.vol_smile_daily_df['expiry_date'].isin(fixing_date_grid)
        #vol_smile_daily_df = self.vol_smile_daily_df.loc[mask].copy()
        self._solve_vol_daily_smile_func(fixing_date_grid)

        # It doesn't really make sense to simulate a rate path using the Heston model.
        # The heston model is designed to for fit the volatility smile at a certain point in time, not the term structure of volatility along the path.

        results = {}

        for i, (fixing_date, delivery_date) in enumerate(zip(fixing_date_grid,delivery_date_grid)):
            print(i, fixing_date, delivery_date)

            expiry_years = year_frac(self.curve_date, fixing_date, self.day_count_basis)
            fx_forward_rate = interp_fx_forward_curve_df(self.fx_forward_curve_df,
                                                         dates=fixing_date,
                                                         date_type='fixing_date').values

            mu = np.log(fx_forward_rate / self.fx_spot_rate) / expiry_years

            var0 = self.vol_smile_daily_func[fixing_date]['var0']
            vv = self.vol_smile_daily_func[fixing_date]['vv']
            kappa = self.vol_smile_daily_func[fixing_date]['kappa']
            theta = self.vol_smile_daily_func[fixing_date]['theta']
            rho = self.vol_smile_daily_func[fixing_date]['rho']
            # lambda_ is not used in the simulation, used in the calibration only.
            # lambda_ = self.daily_volatility_smile_func[fixing_date]['lambda_']

            rand_nbs = generate_rand_nbs(
                nb_steps=100,
                nb_rand_vars=2,
                nb_simulations=10 * 1000,
                apply_antithetic_variates=False)

            sim_results = heston_simulate(
                S0=self.fx_spot_rate,
                mu=mu,
                var0=var0,
                vv=vv,
                kappa=kappa,
                theta=theta,
                rho=rho,
                tau=expiry_years,
                rand_nbs=rand_nbs)

            results[delivery_date] = sim_results

        return results


    def plot_smile(self, expiry_date: pd.Timestamp):

        self._solve_vol_daily_smile_func(pd.DatetimeIndex([expiry_date]))

        vol_daily_smile_df = self.vol_smile_daily_df.loc[self.vol_smile_daily_df['expiry_date'] == expiry_date]

        delta_convention = vol_daily_smile_df['delta_convention'].iloc[0]
        vol_pillar = vol_daily_smile_df[self.quotes_column_names].iloc[0]

        pillar_strikes = garman_kohlhagen_solve_strike_from_delta(S0=self.fx_spot_rate,
                                  tau=vol_daily_smile_df['expiry_years'].iloc[0],
                                  r_d=vol_daily_smile_df['domestic_zero_rate'].iloc[0],
                                  r_f=vol_daily_smile_df['foreign_zero_rate'].iloc[0],
                                  vol=vol_pillar,
                                  signed_delta=self.quotes_signed_delta,
                                  delta_convention=delta_convention,
                                  F=vol_daily_smile_df['fx_forward_rate'].iloc[0],
                                  atm_delta_convention='forward')

        nb_strikes = 100
        strike_delta_neutral = pillar_strikes[self.quotes_column_names == 'atm_delta_neutral'].item()
        strikes_put = np.linspace(np.min(pillar_strikes), strike_delta_neutral, nb_strikes)
        strikes_call = np.linspace(strike_delta_neutral, np.max(pillar_strikes), nb_strikes)
        strikes_array = np.concatenate([strikes_put, strikes_call])
        call_put_array = np.concatenate([-1 * np.ones(nb_strikes), np.ones(nb_strikes)])
        expiry_dates = pd.DatetimeIndex([expiry_date for _ in range(len(strikes_array))])

        result = self.price_vanilla_european(expiry_dates=expiry_dates,
                                             K=strikes_array,
                                             cp=call_put_array,
                                             analytical_greeks_flag=True)

        if delta_convention == 'regular_spot':
            delta_str = 'spot_delta'
        elif delta_convention == 'regular_forward':
            delta_str = 'forward_delta'

        delta = result['analytical_greeks'][delta_str]
        vol = result['market_data_inputs']['vol']

        delta_pillar_for_plot = np.full(self.quotes_signed_delta.shape, np.nan)
        midpoint = len(delta_pillar_for_plot) // 2
        delta_pillar_for_plot[:midpoint] = np.abs(self.quotes_signed_delta[:midpoint])
        delta_pillar_for_plot[midpoint:] = 1.0 - self.quotes_signed_delta[midpoint:]

        delta_for_plot = np.full(delta.shape, np.nan)
        delta_for_plot[:nb_strikes] = np.abs(delta[:nb_strikes])
        delta_for_plot[nb_strikes:] = 1.0 - delta[nb_strikes:]

        # Plot
        plt.plot(100 * delta_pillar_for_plot, vol_pillar, 'x', label='Pillar Volatilities')
        plt.plot(100 * delta_for_plot, vol, label='Fitted Smile')  # X markers

        tenor = np.round(vol_daily_smile_df['expiry_years'].iloc[0], 2)

        plt.title(f'Volatility Smile for {tenor}Y ({expiry_date.date()})\n'
                  f'Smile interpolation method: {self.smile_interpolation_method.value}')
        plt.xlabel('Delta:   Puts: (0,0.5). ATM Δ Neutral: 0.5. Calls: (0.5,1)')
        plt.ylabel('Volatility')
        plt.xlim([0, 100])

        plt.gca().yaxis.set_major_formatter(mtick.PercentFormatter(1.0, decimals=1))  # 1.0 means 100%

        plt.legend()
        plt.show()



# -*- coding: utf-8 -*-
import os
import numpy as np
import warnings

if __name__ == "__main__":
    os.chdir(os.environ.get('PROJECT_DIR_FRM')) 


def forward_volatility(t1: float | np.ndarray,
                       vol_t1: float | np.ndarray,
                       t2: float| np.ndarray,
                       vol_t2: float | np.ndarray) -> float | np.ndarray:
    """
    Calculate forward volatility from time t1 to time t2 using the consistency condition.
    Consistency condition: vol_t1**2 * t1 + vol_t1_t2**2 * (t2 - t1) = vol_t2**2 * t2

    Parameters:
    - t1 (float or np.array): Time to first maturity in years, must be less than t2.
    - vol_t1 (float or np.array): Annualized volatility to expiry at t1 for a given delta.
    - t2 (float or np.array): Time to second maturity in years, must be greater than t1.
    - vol_t2 (float or np.array): Annualized volatility to expiry at t2 for the same delta.

    Returns:
    - float or np.array: Forward volatility from time t1 to t2.

    Raises:
    - ValueError: If t2 < t1 or if a negative value is encountered under the square root.

    Warnings:
    - If t1 and t2 are equal, NaN values are returned for those instances, with a warning.

    Notes:
    - Forward volatility is computed based on the implied variance between times t1 and t2.
    - Negative values under the square root indicate inconsistent input volatilities and will raise an error.
    """

    tau = t2 - t1
    if np.any(tau == 0):
        mask = tau == 0
        warnings.warn(f"t2 and t1 are equal. NaN will be returned for these values: t1 {t1[mask]}, t2 {t2[mask]}")
    elif np.any(tau < 0):
        raise ValueError("t2 is less than t1.")

    var_t1_t2 = (vol_t2 ** 2 * t2 - vol_t1 ** 2 * t1) / tau
    if np.any(var_t1_t2 < 0):
        raise ValueError("Negative value encountered under square root.")

    return np.sqrt(var_t1_t2)


def flat_forward_interp(t1: float | np.ndarray,
                        vol_t1: float | np.ndarray,
                        t2: float | np.ndarray,
                        vol_t2: float | np.ndarray,
                        t: float | np.ndarray) -> float | np.ndarray:
    """
    Interpolate volatility at a specified time 't' using flat forward interpolation.

    Parameters:
    - t1 (float or np.array): Time to first expiry in years, must be less than t2.
    - vol_t1 (float or np.array): Annualized volatility to expiry at t1 for a given delta.
    - t2 (float or np.array): Time to second expiry in years, must be greater than t1.
    - vol_t2 (float or np.array): Annualized volatility to expiry at t2 for the same delta.
    - t (float or np.array): Time at which to interpolate the volatility. Must satisfy t1 <= t <= t2.

    Returns:
    - float or np.array: Interpolated volatility at time 't'.

    Notes:
    - If t is outside the range [t1, t2], the function will not compute interpolation and will use the input boundaries.
    - Assumes a flat forward interpolation model, where volatilities are averaged in a time-weighted manner.
    """
    vol_t12 = np.zeros_like(vol_t1)
    mask = (t1 != t2).flatten()
    vol_t12[mask] = forward_volatility(t1[mask], vol_t1[mask], t2[mask], vol_t2[mask])
    return np.sqrt((vol_t1 ** 2 * t1 + vol_t12 ** 2 * (t - t1)) / t)




# -*- coding: utf-8 -*-

from frm.enums import DayCountBasis

# Global settings
VALUE_DATE = None
INCLUDE_PAYMENTS_ON_VALUE_DATE_IN_NPV  = False
INCLUDE_MONTHS_NON_BUSINESS_DAYS_ACCRUED_INTEREST_IF_VALUE_DATE_IS_LAST_BUSINESS_DAY_OF_MONTH = False
LIMIT_ACCRUED_INTEREST_TO_UNSETTLED_CASHFLOW = False
ROLL_USER_SPECIFIED_DATES  = False

# Day count basis
REPORTING_DAY_COUNT_BASIS = DayCountBasis._30_360 # For current/non-current splits and cashflow bucketing
DATE2YEARFRAC_DAY_COUNT_BASIS = DayCountBasis.ACT_365 #

# Interpolation methods
ZERO_CURVE_INTERPOLATION_METHOD  = 'linear_on_log_of_discount_factors'
# add a few more methods
# - cubic / smoothed spline on ln of discount factors
# - cubic / smoothed  spline on zero rates
# - cubic / smoothed  linear on zero rates
# - flat forward (linear on log of zeros
# https://stackoverflow.com/questions/78054656/what-is-the-difference-between-the-various-spline-interpolators-from-scipy

# MTM CCIRS - notional definition for day 1 or current coupon for single pricer or bulk run

# BBG - IBOR ← → RFR adjustment spreads.

# Method of calculating greeks. (analycal, bump and reprice)


INFLATION_FORWARD_INTERPOLATION_METHOD  = 'linear' # 'exponential'


# Lyashenko, A., & Mercurio, F. (2019). Looking Forward to Backward-Looking Rates: A Modeling Framework for Term Rates Replacing LIBOR. SSRN Electronic Journal. doi:10.2139/ssrn.3330240
RFR_OPTIONLET_EXPIRY_METHOD = ['date_of_last_fixing','3rd_way_through_period']

# Swaption
# Cash settlement method - 'market default annuity' or 'swap curve annuity'.

# Greeks
vega_normalisation = +0.01 # +1%
dv01_adjustment = +0.0001 # +0.01%

# Excel formats
EXCEL_DATE_FORMAT = '%d-%b-%Y'
EXCEL_AMT_FORMAT = '#,##0_-;(#,##0)_-;-_-'


MAX_SIMULATIONS_PER_LOOP = 100_000_000  # Maximum number of total random numbers per loop

# Default number of monte carlo paths

# extroplotation methods
# flat on rate / yield
# do not apply





# -*- coding: utf-8 -*-
import os
if __name__ == "__main__":
    os.chdir(os.environ.get('PROJECT_DIR_FRM')) 
        
import numpy as np
from frm.enums import CompoundingFreq


def zero_rate_from_discount_factor(
        years, 
        discount_factor, 
        compounding_freq: CompoundingFreq):
    
    if compounding_freq == CompoundingFreq.CONTINUOUS:
            return (- np.log(discount_factor) / years)    
    elif compounding_freq == CompoundingFreq.SIMPLE:
        return (((1.0 / discount_factor) - 1.0) / years)       
    elif compounding_freq in {
            CompoundingFreq.DAILY,
            CompoundingFreq.MONTHLY,
            CompoundingFreq.QUARTERLY,
            CompoundingFreq.SEMIANNUAL,
            CompoundingFreq.ANNUAL}:
        periods_per_year = compounding_freq.periods_per_year
        return periods_per_year * ((1.0 / discount_factor) ** (1.0 / (periods_per_year * years)) - 1.0)
    else:
        raise ValueError(f"Invalid compounding_frequency {compounding_freq}")
    

def discount_factor_from_zero_rate(
        years, 
        zero_rate, 
        compounding_freq: CompoundingFreq):
    
    if compounding_freq == CompoundingFreq.CONTINUOUS:
        return np.exp(-zero_rate * years)
    if compounding_freq == CompoundingFreq.SIMPLE:
        return  1.0 / (1.0 + zero_rate * years)
    elif compounding_freq in {
            CompoundingFreq.DAILY,
            CompoundingFreq.MONTHLY,
            CompoundingFreq.QUARTERLY,
            CompoundingFreq.SEMIANNUAL,
            CompoundingFreq.ANNUAL}:
        periods_per_year = compounding_freq.periods_per_year
        return 1.0 / (1.0 + zero_rate / periods_per_year) ** (periods_per_year * years)
    else:
        raise ValueError(f"Invalid compounding_frequency {compounding_freq}")
        
        

            

            
            
            
            
            
            
            
            
            
            


# -*- coding: utf-8 -*-
import os
import numpy as np
from numpy.typing import NDArray
import scipy
from typing import Optional
import numbers

if __name__ == "__main__":
    os.chdir(os.environ.get('PROJECT_DIR_FRM'))


def calc_sln_vol_for_strike_from_sabr_params(
        tau: [float, NDArray[float]],
        F: [float, NDArray[float]],
        alpha: [float, NDArray[float]],
        beta: [float, NDArray[float]],
        rho: [float, NDArray[float]],
        volvol: [float, NDArray[float]],
        K: [float, NDArray[float]],
        ln_shift: float) -> [float, NDArray[float]]:
    """
    Calculate the log-normal (Black76) volatility for a given strike price using SABR model parameters.

    This implementation follows the Hagan et al. (2002) model for managing smile risk, applying formulas 2.17 and 2.18.

    Parameters:
    -----------
    tau : float or array-like of float
        Time to expiry in years. Must be either scalar or 1D array matching the dimensions of `F`, `alpha`, etc.
    F : float or array-like of float
        Forward rate, scalar or 1D array.
    alpha : float or array-like of float
        Initial volatility level, scalar or 1D array.
    beta : float or array-like of float
        Elasticity of the volatility with respect to the forward rate, scalar or 1D array.
    rho : float or array-like of float
        Correlation between the forward rate and the volatility, scalar or 1D array.
    volvol : float or array-like of float
        Volatility of volatility, scalar or 1D array.
    K : float or array-like of float
        Strike price, scalar or 1D array.
    ln_shift : float
        Log-normal shift to adjust both `F` and `K` to prevent zero or near-zero log values.

    Returns:
    --------
    float or ndarray
        Log-normal volatility for the given strike price(s). Returns a scalar if all inputs are scalars, otherwise returns an ndarray.

    Notes:
    ------
    If the `F` and `K` values are very close (at-the-money condition), an alternate formula (2.18) is applied for higher accuracy.

    References:
    -----------
    [1] Hagan, P., Lesniewski, A., & Woodward, D. (2002). Managing Smile Risk. Wilmott Magazine, 1, 84-108.
    """

    tau, F, alpha, beta, rho, volvol, K, ln_shift = map(np.atleast_1d, (tau, F, alpha, beta, rho, volvol, K, ln_shift))

    # Input must be 1D column vector
    for param in (tau, F, alpha, beta, rho, volvol):
        assert param.ndim == 1, 'Parameters must be 1D arrays or 2D column vectors with shape (n, 1).'

    assert tau.shape == F.shape == alpha.shape == beta.shape == rho.shape == volvol.shape, \
        'tau, F, alpha, beta, rho, volvol must have the same shape as they define the SABR smile for a given expiry.'

    if K.size > 1:
        assert tau.size == K.size or tau.size == 1, \
            'K must be the same shape as the SABR smile parameters (tau, F, alpha, beta, rho, volvol), and/or a scalar, (1,).'

    if ln_shift.size == 1:
        ln_shift = ln_shift.item()
    assert isinstance(ln_shift, numbers.Real), 'ln_shift must be a valid numeric scalar.'


    F = F + ln_shift
    K = K + ln_shift

    # Set to symbols used in [1] for easier comparison
    α = alpha
    β = beta
    ρ = rho
    v = volvol

    mask_atm = np.abs(np.log(F / K)) < 1e-06
    σB = np.full(mask_atm.shape, np.nan)

    if mask_atm.any():
        # If ATM, use (2.18) in [1].
        # At ATM, F=K which simplifies the formulae in 2.17.
        σB[mask_atm] = (α /
                (F*K)**((1 - β) / 2) *
                (1
                + ( ((1-β)**2 / 24) * α**2 / ((F*K)**(1-β))
                    + (0.25 * ρ * β * v * α / ((F*K)**((1-β)/2)))
                    + ((2 - 3 * ρ**2) * v**2 / 24)
                ) * tau))[mask_atm]

    if ~mask_atm.all():
        # (2.17b) in [1]
        z = (v / α) * (F * K)**((1 - β) / 2) * np.log(F / K)

        def x(z):
            # (2.17c) in [1]
            return np.log((np.sqrt(1 - 2 * ρ * z + z ** 2) + z - ρ) / (1 - ρ))

        # Row 1 in (2.17a) in [1]
        row1 = α * (z / x(z)) / (
                (F*K)**((1 - β) / 2) *
                (1
                 + (1 - β)**2 / 24 * np.log(F / K)**2
                 + (1 - β)**4 / 1920 * np.log(F / K)**4))

        # Row 2 in (2.17a) in [1]
        row2 = (1
                + ( ((1-β)**2 / 24) * (α**2 / ((F*K)**(1-β)))
                    + (0.25 * ρ * β * v * α / ((F*K)**((1-β)/2)))
                    + ((2 - 3 * ρ**2) * v**2 / 24)
                ) * tau)

        σB[~mask_atm] = (row1 * row2)[~mask_atm]

    if σB.size == 1:
        return σB.item()
    else:
        return σB


def solve_alpha_from_sln_vol(
    tau: float,
    F: float,
    beta: float,
    rho: float,
    volvol: float,
    vol_sln_atm: float,
    ln_shift: float = 0.0) -> float:
    """
    Solve alpha analytically by rearranging (2.18) in [1] for alpha.
    The equation is a 3rd degree polynomial (in alpha).

    Parameters:
    ----------
    tau: float
        Time to maturity in years.
    F: float
        Forward rate.
    beta: float
        Elasticity of the volatility with respect to the forward rate.
    rho: float
        Correlation between the forward rate and the volatility.
    volvol: float
        Volatility of volatility.
    vol_sln_atm: float
        ATMF log-normal volatility.
    ln_shift: float
        Log-normal shift.

    Returns:
    --------
    float
        Solved alpha.

    References:
    [1] Hagan, Patrick & Lesniewski, Andrew & Woodward, Diana. (2002). Managing Smile Risk. Wilmott Magazine. 1. 84-108.
    """

    F = F + ln_shift
    β, ρ, v = beta, rho, volvol
    del beta, rho, volvol

    # Rearrange equation 2.18 from Hagan 2002 into cubic form: Aα^3 + Bα^2 + Cα + K = 0
    A = (1/24) * ((1-β)**2) * tau / (F**(2-2*β))
    B = (1/4) * ρ * β * v * tau / (F**(1-β))
    C = (1 + (2-3*ρ**2) * v**2 * tau/ 24)
    K =  -vol_sln_atm * F**(1-β)
    cubic_polynomial = [A, B, C, K]

    roots = np.roots(cubic_polynomial)
    roots_real = np.extract(np.isreal(roots), np.real(roots))

    # Note: the double real roots case is not tested
    alpha_first_guess = vol_sln_atm * F**(1-β)
    i_min = np.argmin(np.abs(roots_real - alpha_first_guess))
    return roots_real[i_min].item()


def fit_sabr_params_to_sln_smile(tau: float,
                                 F: float,
                                 K: NDArray[float],
                                 vols_sln: NDArray[float],
                                 ln_shift: float=None,
                                 beta_overide: Optional[float]=None)-> tuple:
    """
    Calibrates SABR parameters to a volatility smile of European option prices using the Hagan 2002 (lognormal) SABR model and the Black76 formula.
    This method is not suitable for caps/floors or swaptions.

    Notes:
    The rho, volvol and (optionally) beta parameters are solved numerically.
    The beta parameter is optional as it is common for this parameter to be fixed by the practitioner.
    The alpha parameter is solved analytically from the other SABR parameters and the ATM volatility.

    Parameters:
    -----------
    tau: float
        Time to maturity in years.
    F: float
        Forward rate.
    ln_shift: float
        Log-normal shift.
    K: np.array
        Strike prices.
    vols_sln: np.array
        (shifted) log-normal volatilities of European options.
    beta: Optional[float]
        Elasticity of the volatility with respect to the forward rate.

    Returns:
    --------
    tuple
        SABR parameters (alpha, beta, rho, volvol), scipy.optimize.OptimizeResult
    """

    def vol_sse(param):
        if beta_overide is None:
            beta_, rho_, volvol_ = param
        else:
            rho_, volvol_ = param
            beta_ = beta_overide

        alpha_ = solve_alpha_from_sln_vol(tau=tau, F=F, beta=beta_, rho=rho_, volvol=volvol_, vol_sln_atm=vol_sln_atm, ln_shift=ln_shift)
        sabr_vols = calc_sln_vol_for_strike_from_sabr_params(tau=tau, F=F, alpha=alpha_, beta=beta_, rho=rho_, volvol=volvol_ ,K=K, ln_shift=ln_shift)
        return sum((vols_sln - sabr_vols) ** 2)

    # ATM volatility check
    vol_sln_atm = vols_sln[np.isclose(K, F)]
    if vol_sln_atm.size != 1:
        raise ValueError('A unique ATM strike must be present.')
    vol_sln_atm = np.atleast_1d(vol_sln_atm).item()


    # params = (beta), rho, volvol
    # beta has a valid range of 0≤β≤1
    # rho has a valid range of -1≤ρ≤1
    # volvol has a valid range of 0<v≤∞
    x0 = np.array([0.00, 0.10]) if beta_overide is not None else np.array([0.0, 0.0, 0.1])
    bounds = [(-1.0, 1.0), (0.0001, None)] if beta_overide is not None else [(-1.0, 1.0), (-1.0, 1.0), (0.0001, None)]
    res = scipy.optimize.minimize(fun=lambda param: vol_sse(param), x0=x0, bounds=bounds)

    if res.success:
        beta, rho, volvol = (beta_overide, *res.x) if beta_overide is not None else res.x
        alpha = solve_alpha_from_sln_vol(tau=tau, F=F, beta=beta, rho=rho, volvol=volvol, vol_sln_atm=vol_sln_atm, ln_shift=ln_shift)
        return (alpha, beta, rho, volvol), res
    else:
        print(res)
        raise ValueError(f'Optimization of SABR parameters failed.')








# -*- coding: utf-8 -*-
from abc import ABC, abstractmethod
from dataclasses import dataclass, field, InitVar
from enum import Enum
import numpy as np
import os
import pandas as pd
from typing import Optional, Union
import warnings

import frm.utils
from frm.utils import CouponSchedule, day_count, year_frac, get_busdaycal
from frm.enums import CompoundingFreq, TermRate, RFRFixingCalcMethod, PeriodFreq, DayCountBasis, ExchangeNotionals
from frm.term_structures.zero_curve import ZeroCurve
from frm.term_structures.zero_curve_helpers import discount_factor_from_zero_rate
from scipy.optimize import root_scalar


if __name__ == "__main__":
    os.chdir(os.environ.get('PROJECT_DIR_FRM'))

# I am writing a (python) class to support the valuation of generic swap legs and bonds.

# THe key methods for a 'leg' are
# (i) coupon date schedule → schedule with forward coupon cashflow amount (will be different for each leg type)
# (ii) discount cashflows & get PV (this will be identical for all leg types)
# (iii) solve parameter (coupon rate, spread) in order to match leg to given price/value.
# (iv) clean mv, dirty mv, accrued interest, cashflow buckets, current, non-current, etc.
# (v) DV01, +/- 100 bps shift to par rates

# Leg types:
# Fixed, Zerocoupon
# FloatTerm, FloatRFR, InflationZC, InflationYoY
# Each leg type will have a different method for calculating the forward coupon cashflow amount.





class PayRcv(Enum):
    PAY = 'pay'
    RCV = 'rcv'

    @property
    def multiplier(self):
        return -1 if self == PayRcv.PAY else 1

@dataclass
class Leg(ABC):
    "Generic class to cover pricing of all types of swap legs and bonds"

    # Required parameters
    schedule: CouponSchedule
    discount_curve: Optional[ZeroCurve]
    pay_rcv: Optional[PayRcv]=PayRcv.PAY
    day_count_basis: DayCountBasis=DayCountBasis

    # Optional parameters
    notional_ccy: str=None
    settlement_ccy: str=None

    def __post_init__(self):
        self.schedule.df.loc[:,['notional', 'notional_payment']] = self.schedule.df[['notional', 'notional_payment']] * self.pay_rcv.multiplier
        self.schedule.add_period_daycount(day_count_basis=self.day_count_basis)
        self.schedule.add_period_yearfrac(day_count_basis=self.day_count_basis)

    @abstractmethod
    def calc_coupon_payment(self, coupon_override=None):
        pass

    @abstractmethod
    def _calc_accrued_interest(self):
        pass

    @abstractmethod
    def _get_solved_field_name(self):
        pass

    @abstractmethod
    def _get_legtype_specific_schedule_cols(self):
        pass

    def _get_contractual_schedule_cols(self):

        abstract_columns = ['period_start', 'period_end', 'notional_payment_date', 'coupon_payment_date']
        legtype_specific_columns = self._get_legtype_specific_schedule_cols()

        cols = []
        for col in self.schedule.df.columns:
            if col in abstract_columns or col in legtype_specific_columns:
                cols.append(col)
        return cols

    def target_solve(self, target_value):
        solved_value = self._solver_helper(solved_fieldname=self._get_solved_field_name(), target_value=target_value)
        return solved_value

    def par_solve(self):

        if 'notional_payment_date' not in self.schedule.df.columns:
            par_value = 0
        else:
            if frm.utils.INCLUDE_PAYMENTS_ON_VALUE_DATE_IN_NPV:
                mask = self.schedule.df['notional_payment_date'] >= frm.utils.VALUE_DATE
            else:
                mask = self.schedule.df['notional_payment_date'] > frm.utils.VALUE_DATE
            # Par value is the outstanding notional payments.
            par_value = self.schedule.df['notional_payment'][mask].sum()

        solved_value = self._solver_helper(solved_fieldname=self._get_solved_field_name(), target_value=par_value)
        return solved_value

    def _solver_helper(self, solved_fieldname, target_value):
        field_original_value = self.schedule.df[solved_fieldname].values.copy()

        def objective_function(overide):
            return self.calc_dirty_pv(coupon_override=overide) - target_value

        solved_value = np.nan
        try:
            bracket = [-0.1, 3] # -10% to 300%
            solution = root_scalar(objective_function, bracket=bracket, method='brentq')
            if solution.converged:
                solved_value = solution.root
        finally:
            # Ensure reset, even if the solver fails
            self.schedule.df[solved_fieldname] = field_original_value

            return solved_value

    def calc_accrued_interest(self):
        # Leg specific accrued interest calculation
        accrued_interest = self._calc_accrued_interest()

        # Calculate the unsettled cashflows
        if frm.utils.INCLUDE_PAYMENTS_ON_VALUE_DATE_IN_NPV:
            mask = (self.schedule.df['coupon_payment_date'] >= frm.utils.VALUE_DATE) & (self.schedule.df['period_end'] <= frm.utils.VALUE_DATE)
        else:
            mask = (self.schedule.df['coupon_payment_date'] > frm.utils.VALUE_DATE) & (self.schedule.df['period_end'] <= frm.utils.VALUE_DATE)
        if mask.sum() > 1:
            warnings.warn('Multiple periods found for unsettled cashflow calculation.')
        unsettled_cashflows = self.schedule.df['coupon_payment'].sum()

        if frm.utils.LIMIT_ACCRUED_INTEREST_TO_UNSETTLED_CASHFLOW:
            if unsettled_cashflows > 0:
                return unsettled_cashflows
            else:
                return accrued_interest
        else:
            return accrued_interest + unsettled_cashflows

    def _discount_payments(self, prefix=''):
        payment_date_col = prefix + 'payment_date'
        discount_factor_col = prefix + 'discount_factor'

        if payment_date_col in self.schedule.df.columns:
            dates = self.schedule.df[payment_date_col]
            self.schedule.df[discount_factor_col] = self.discount_curve.get_discount_factors(self.schedule.df[payment_date_col])
            self.schedule.df.loc[pd.isna(dates),discount_factor_col] = 0.0

        if frm.utils.INCLUDE_PAYMENTS_ON_VALUE_DATE_IN_NPV:
            mask = self.schedule.df[payment_date_col] < frm.utils.VALUE_DATE
        else:
            mask = self.schedule.df[payment_date_col] <= frm.utils.VALUE_DATE
        self.schedule.df.loc[mask, discount_factor_col] = 0.0

    def calc_dirty_pv(self, coupon_override=None):
        self.calc_coupon_payment(coupon_override=coupon_override)
        self.schedule.df['total_payment'] = self.schedule.df['notional_payment'] + self.schedule.df['coupon_payment']
        self._discount_payments(prefix='notional_')
        self._discount_payments(prefix='coupon_')
        self.schedule.df['notional_payment_pv'] = (self.schedule.df['notional_payment'] *
                                                               self.schedule.df['notional_discount_factor'])
        self.schedule.df['coupon_payment_pv'] = (self.schedule.df['coupon_payment'] *
                                                            self.schedule.df['coupon_discount_factor'])
        self.schedule.df['total_payment_pv'] = (self.schedule.df['notional_payment_pv'] +
                                                           self.schedule.df['coupon_payment_pv'])
        return self.schedule.df['total_payment_pv'].sum()

    def calc_clean_pv(self):
        return self.calc_dirty_pv() - self.calc_accrued_interest()

    def get_current_non_current_split(self, calc_npv=True):
        if calc_npv:
            self.calc_dirty_pv()

        current, non_current = 0, 0
        value_date_plus_one_year = frm.utils.VALUE_DATE + pd.DateOffset(years=1)

        if frm.utils.INCLUDE_PAYMENTS_ON_VALUE_DATE_IN_NPV:
            current_period_idxs = self.schedule.df['notional_payment_date'] <= value_date_plus_one_year
        else:
            current_period_idxs = self.schedule.df['notional_payment_date'] < value_date_plus_one_year
        current += self.schedule.df['notional_payment_pv'][current_period_idxs].sum()
        non_current += self.schedule.df['notional_payment_pv'][~current_period_idxs].sum()

        if frm.utils.INCLUDE_PAYMENTS_ON_VALUE_DATE_IN_NPV:
            current_period_idxs = self.schedule.df['coupon_payment_date'] <= value_date_plus_one_year
        else:
            current_period_idxs = self.schedule.df['coupon_payment_date'] < value_date_plus_one_year
        current += self.schedule.df['coupon_payment_pv'][current_period_idxs].sum()
        non_current += self.schedule.df['notional_payment_pv'][~current_period_idxs].sum()

        return current, non_current

    def _bucket_payments_helper(self, buckets, prefix):
        payment_col = prefix + 'payment'
        payment_date_col = prefix + 'payment_date'

        l = []
        if frm.utils.INCLUDE_PAYMENTS_ON_VALUE_DATE_IN_NPV:
            mask_lower = self.schedule.df[payment_date_col] >= frm.utils.VALUE_DATE
        else:
            mask_lower = self.schedule.df[payment_date_col] > frm.utils.VALUE_DATE

        for i,v in enumerate(buckets):
            upper_pillar = frm.utils.VALUE_DATE + pd.DateOffset(years=v)
            mask_upper = self.schedule.df[payment_date_col] < (upper_pillar)
            l.append(self.schedule.df[payment_col][mask_lower & mask_upper].sum())
            mask_lower = self.schedule.df[payment_date_col] >= upper_pillar

        mask_lower = self.schedule.df[payment_date_col] >= upper_pillar
        l.append(self.schedule.df[payment_col][mask_lower].sum())
        index = buckets + (np.inf,)

        return pd.DataFrame(l, index=index, columns=[payment_col])

    def bucket_payments(self, calc_npv=True, buckets=(1,2,3,5)):
        if calc_npv:
            self.calc_dirty_pv()
        coupon_payments_bucketed = self._bucket_payments_helper(buckets, prefix='coupon_')
        notional_payments_bucketed = self._bucket_payments_helper(buckets, prefix='notional_')
        combined = pd.concat([coupon_payments_bucketed, notional_payments_bucketed], axis=1)
        combined['total_payment'] = combined['coupon_payment'] + combined['notional_payment']
        return combined


    def calc_dv01(self):
        # Shared DV01 calculation with +/- 100 bps shift
        pass



@dataclass
class FixedLeg(Leg):
    fixed_rate: float=np.nan

    def __post_init__(self):
        super().__post_init__()
        self.schedule.df['fixed_rate'] = self.fixed_rate

    def calc_coupon_payment(self, coupon_override=None):
        if coupon_override is not None:
            self.schedule.df['fixed_rate'] = coupon_override
        self.schedule.df['coupon_payment'] = self.pay_rcv.multiplier * self.schedule.df['notional'] * self.schedule.df['fixed_rate'] * self.schedule.df['period_yearfrac']
        mask = pd.isna(self.schedule.df['coupon_payment_date'])
        self.schedule.df.loc[mask,'coupon_payment'] = 0.0

    def _calc_accrued_interest(self):
        current_period_idxs = (self.schedule.df['period_start'] <= frm.utils.VALUE_DATE) & (self.schedule.df['period_end'] > frm.utils.VALUE_DATE)
        if current_period_idxs.sum() > 1:
            warnings.warn('Multiple periods found for accrual interest calculation.')
        accrued_period_yearfrac = year_frac(self.schedule.df['period_start'][current_period_idxs].iloc[0],
                                         frm.utils.VALUE_DATE,
                                         day_count_basis=self.day_count_basis)
        accrued_interest = (self.schedule.df['notional'][current_period_idxs]
                            * self.schedule.df['fixed_rate'][current_period_idxs]
                            * accrued_period_yearfrac).sum()
        return accrued_interest

    def _get_solved_field_name(self):
        return 'fixed_rate'

    def _get_legtype_specific_schedule_cols(self):
        return ['fixed_rate']


@dataclass
class FloatTermLeg(Leg):
    spread: float=np.nan
    forward_rate_type: TermRate=TermRate.SIMPLE
    forward_curve: Optional[ZeroCurve]=None

    def __post_init__(self):
        super().__post_init__()
        self.schedule.df[self._get_legtype_specific_schedule_cols()] = np.nan
        self.schedule.df['spread'] = self.spread
        #self.calc_notional_schedule(coupon_fields_to_set_to_zero=['fixing','spread','coupon_rate'])

    def calc_coupon_payment(self, coupon_override=None):
        self.schedule.df['fixing'] = np.nan
        if coupon_override is not None:
            self.schedule.df['spread'] = coupon_override
        self.schedule.df['coupon_rate'] = self.schedule.df['fixing'] + self.schedule.df['spread']
        self.schedule.df['coupon_payment'] = self.schedule.df['notional'] * self.schedule.df['coupon_rate'] * self.schedule.df['period_yearfrac']
        mask = pd.isna(self.schedule.df['coupon_payment_date'])
        self.schedule.df.loc[mask,'coupon_payment'] = 0.0

    def _calc_accrued_interest(self):
        current_period_idxs = (self.schedule.df['period_start'] <= frm.utils.VALUE_DATE) & (self.schedule.df['period_end'] > frm.utils.VALUE_DATE)
        if current_period_idxs.sum() > 1:
            warnings.warn('Multiple periods found for accrual interest calculation.')
        accrued_period_yearfrac = year_frac(self.schedule.df['period_start'][current_period_idxs].iloc[0], VALUE_DATE, day_count_basis=self.day_count_basis)
        accrued_interest = (self.schedule.df['notional'][current_period_idxs]
                            * self.schedule.df['spread'][current_period_idxs]
                            * accrued_period_yearfrac).sum()
        return accrued_interest

    def _get_solved_field_name(self):
        return 'spread'

    def _get_legtype_specific_schedule_cols(self):
        return ['fixing_date','fixing','spread','coupon_rate']


@dataclass
class FloatRFRLeg(Leg):
    spread: float=np.nan
    forward_rate_type: RFRFixingCalcMethod=RFRFixingCalcMethod.DAILY_COMPOUNDED
    compound_spread: bool=False # TODO
    forward_curve: Optional[ZeroCurve]=None

    def __post_init__(self):
        super().__post_init__()
        self.schedule.df[self._get_legtype_specific_schedule_cols()] = np.nan
        self.schedule.df['spread'] = self.spread
        #self.calc_notional_schedule(coupon_fields_to_set_to_zero=['fixing','spread','coupon_rate'])

    def calc_coupon_payment(self, coupon_override=None):
        self.schedule.df['fixing'] = np.nan # TODO
        if coupon_override is not None:
            self.schedule.df['spread'] = coupon_override
        self.schedule.df['coupon_rate'] = self.schedule.df['fixing'] + self.schedule.df['spread']
        self.schedule.df['coupon_payment'] = self.schedule.df['notional'] * self.schedule.df['coupon_rate'] * self.schedule.df['period_yearfrac']
        mask = pd.isna(self.schedule.df['coupon_payment_date'])
        self.schedule.df.loc[mask,'coupon_payment'] = 0.0

    def _calc_accrued_interest(self):
        current_period_idxs = (self.schedule.df['period_start'] <= frm.utils.VALUE_DATE) & (self.schedule.df['period_end'] > frm.utils.VALUE_DATE)
        if current_period_idxs.sum() > 1:
            warnings.warn('Multiple periods found for accrual interest calculation.')
        accrued_period_yearfrac = year_frac(self.schedule.df['period_start'][current_period_idxs].iloc[0], frm.utils.VALUE_DATE, day_count_basis=self.day_count_basis)
        accrued_interest = (self.pay_rcv.multiplier * self.schedule.df['notional'][current_period_idxs]
                            * self.schedule.df['coupon_rate'][current_period_idxs]
                            * accrued_period_yearfrac).sum()
        return accrued_interest

    def _get_solved_field_name(self):
        return 'spread'

    def _get_legtype_specific_schedule_cols(self):
        return ['fixing','spread','coupon_rate']


@dataclass
class ZerocouponLeg(Leg):
    zero_rate: float=np.nan
    compounding_freq: CompoundingFreq=CompoundingFreq.ANNUAL

    def __post_init__(self):
        super().__post_init__()
        self.schedule.df['zero_rate'] = self.zero_rate
        # TODO support specifying the terminal notional

    def calc_coupon_payment(self, coupon_override=None):
        if coupon_override is not None:
            self.schedule.df['zero_rate'] = coupon_override
        multiplier = (1.0 / discount_factor_from_zero_rate(
            years=self.schedule.df['period_yearfrac'],
            zero_rate=self.schedule.df['zero_rate'],
            compounding_freq=self.compounding_freq)) - 1.0
        self.schedule.df['coupon_payment'] = self.pay_rcv.multiplier * self.schedule.df['notional'] * multiplier
        mask = pd.isna(self.schedule.df['coupon_payment_date'])
        self.schedule.df.loc[mask,'coupon_payment'] = 0.0

    def _calc_accrued_interest(self):
        current_period_idxs = (self.schedule.df['period_start'] <= frm.utils.VALUE_DATE) & (
                    self.schedule.df['period_end'] > frm.utils.VALUE_DATE)
        if current_period_idxs.sum() > 1:
            warnings.warn('Multiple periods found for accrual interest calculation.')
        accrued_period_yearfrac = year_frac(self.schedule.df['period_start'][current_period_idxs].iloc[0], frm.utils.VALUE_DATE,
                                         day_count_basis=self.day_count_basis)
        multiplier = (1.0 / discount_factor_from_zero_rate(
            years=accrued_period_yearfrac,
            zero_rate=self.schedule.df['zero_rate'],
            compounding_freq=self.compounding_freq)) - 1.0
        accrued_interest =  self.pay_rcv.multiplier * self.schedule.df['notional'] * multiplier
        return accrued_interest

    def _get_solved_field_name(self):
        return 'zero_rate'

    def _get_legtype_specific_schedule_cols(self):
        return ['zero_rate']

# TBC at later date.
class InflationZCLeg(Leg):
    fixing_lag_months: int=2
    initial_fixing: float=np.nan

    def __post_init__(self):
        super().__post_init__()

    def calc_coupon_cashflow(self):
        pass

    def _calc_accrued_interest(self):
        pass

    def _get_solved_field_name(self):
        return 'final_fixing'

# TBC at later date.
class InflationYoYLeg(Leg):
    fixing_lag_months: int=2
    initial_fixing: float=np.nan
    # forward_rate_type: TBC

    def __post_init__(self):
        super().__post_init__()

    def calc_coupon_cashflow(self):
        pass

    def _calc_accrued_interest(self):
        pass

    def _get_solved_field_name(self):
        return 'spread'


